import numpy as np
import unittest

import porepy as pp

from porepy.numerics.vem import hybrid


def make_dictionary(g, perm, bc, bc_val=None):
    if bc_val is None:
        bc_val = np.zeros(g.num_faces)
    d = {
        "bc": bc,
        "bc_values": bc_val,
        "second_order_tensor": perm,
        "aperture": np.ones(g.num_cells),
        "source": np.zeros(g.num_cells),
    }

    return pp.initialize_default_data(g, {}, "flow", d)


class BasicsTest(unittest.TestCase):
    def test_dual_hybrid_vem_1d_iso(self):
        g = pp.CartGrid(3, 1)
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx)
        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["neu"])

        solver = hybrid.HybridDualVEM(keyword="flow")

        data = make_dictionary(g, perm, bc)
        M = solver.matrix_rhs(g, data)[0].todense()

        M_known = 3 * np.array(
            [[-1, 1, 0, 0], [1, -2, 1, 0], [0, 1, -2, 1], [0, 0, 1, -1]]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    def test_dual_hybrid_vem_1d_ani(self):
        g = pp.CartGrid(3, 1)
        g.compute_geometry()

        kxx = np.sin(g.cell_centers[0, :]) + 1
        perm = pp.SecondOrderTensor(kxx)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["neu"])

        solver = hybrid.HybridDualVEM(keyword="flow")

        data = make_dictionary(g, perm, bc)
        M = solver.matrix_rhs(g, data)[0].todense()

        M_known = np.array(
            [
                [-3.4976883980802449, 3.4976883980802453, 0, 0],
                [3.4976883980802453, -7.9359650138928544, 4.4382766158126081, 0],
                [0, 4.4382766158126081, -9.6588071754007174, 5.2205305595881093],
                [0, 0, 5.2205305595881093, -5.2205305595881093],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    def test_dual_hybrid_vem_2d_iso_cart(self):
        g = pp.CartGrid([2, 1], [1, 1])
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["neu"])

        solver = hybrid.HybridDualVEM(keyword="flow")

        data = make_dictionary(g, perm, bc)
        M = solver.matrix_rhs(g, data)[0].todense()

        M_known = np.array(
            [
                [-2.25, 1.75, 0.0, 0.25, 0.0, 0.25, 0.0],
                [1.75, -4.5, 1.75, 0.25, 0.25, 0.25, 0.25],
                [0.0, 1.75, -2.25, 0.0, 0.25, 0.0, 0.25],
                [0.25, 0.25, 0.0, -0.75, 0.0, 0.25, 0.0],
                [0.0, 0.25, 0.25, 0.0, -0.75, 0.0, 0.25],
                [0.25, 0.25, 0.0, 0.25, 0.0, -0.75, 0.0],
                [0.0, 0.25, 0.25, 0.0, 0.25, 0.0, -0.75],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    def test_dual_hybrid_vem_2d_ani_cart(self):
        g = pp.CartGrid([2, 1], [1, 1])
        g.compute_geometry()

        kxx = np.square(g.cell_centers[1, :]) + 1
        kyy = np.square(g.cell_centers[0, :]) + 1
        kxy = -np.multiply(g.cell_centers[0, :], g.cell_centers[1, :])
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kyy, kxy=kxy)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["neu"])

        solver = hybrid.HybridDualVEM(keyword="flow")

        data = make_dictionary(g, perm, bc)
        M = solver.matrix_rhs(g, data)[0].todense()

        M_known = np.array(
            [
                [
                    -2.7386363636363646,
                    2.2613636363636376,
                    0.0,
                    0.3636363636363635,
                    0.0,
                    0.1136363636363635,
                    0.0,
                ],
                [
                    2.2613636363636371,
                    -5.472507331378301,
                    2.2661290322580658,
                    0.1136363636363638,
                    0.6088709677419353,
                    0.3636363636363638,
                    -0.1411290322580648,
                ],
                [
                    0.0,
                    2.2661290322580654,
                    -2.7338709677419368,
                    0.0,
                    -0.1411290322580644,
                    0.0,
                    0.6088709677419357,
                ],
                [
                    0.3636363636363636,
                    0.1136363636363638,
                    0.0,
                    -0.7698863636363639,
                    0.0,
                    0.2926136363636364,
                    0.0,
                ],
                [
                    0.0,
                    0.6088709677419354,
                    -0.1411290322580645,
                    0.0,
                    -1.0151209677419353,
                    0.0,
                    0.5473790322580644,
                ],
                [
                    0.1136363636363636,
                    0.3636363636363638,
                    0.0,
                    0.2926136363636364,
                    0.0,
                    -0.7698863636363638,
                    0.0,
                ],
                [
                    0.0,
                    -0.1411290322580647,
                    0.6088709677419355,
                    0.0,
                    0.5473790322580643,
                    0.0,
                    -1.0151209677419353,
                ],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    def test_dual_hybrid_vem_2d_iso_simplex(self):
        g = pp.StructuredTriangleGrid([1, 1], [1, 1])
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["neu"])

        solver = hybrid.HybridDualVEM(keyword="flow")

        data = make_dictionary(g, perm, bc)
        M = solver.matrix_rhs(g, data)[0].todense()

        M_known = np.array(
            [
                [-2, 0, 2, 0, 0],
                [
                    0,
                    -1.9999999999999996e00,
                    1.9999999999999991e00,
                    0,
                    4.9960036108132044e-16,
                ],
                [
                    2,
                    1.9999999999999993e00,
                    -7.9999999999999982e00,
                    2,
                    1.9999999999999993e00,
                ],
                [
                    -1.1102230246251565e-16,
                    0.0000000000000000e00,
                    2.0000000000000004e00,
                    -2.0000000000000000e00,
                    0,
                ],
                [
                    0,
                    2.2204460492503131e-16,
                    1.9999999999999996e00,
                    0,
                    -1.9999999999999998e00,
                ],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        # We test only the mass-Hdiv part
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    def test_dual_hybrid_vem_2d_ani_simplex(self):
        g = pp.StructuredTriangleGrid([1, 1], [1, 1])
        g.compute_geometry()

        kxx = np.square(g.cell_centers[1, :]) + 1
        kyy = np.square(g.cell_centers[0, :]) + 1
        kxy = -np.multiply(g.cell_centers[0, :], g.cell_centers[1, :])
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kyy, kxy=kxy)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["neu"])

        solver = hybrid.HybridDualVEM(keyword="flow")

        data = make_dictionary(g, perm, bc)
        M = solver.matrix_rhs(g, data)[0].todense()

        M_known = np.array(
            [
                [-2.888888888888888, 0.0, 3.3333333333333326, -0.4444444444444444, 0.0],
                [0.0, -2.888888888888888, 3.3333333333333321, 0.0, -0.4444444444444441],
                [
                    3.3333333333333326,
                    3.3333333333333321,
                    -11.9999999999999964,
                    2.666666666666667,
                    2.6666666666666656,
                ],
                [-0.4444444444444447, 0.0, 2.666666666666667, -2.2222222222222223, 0.0],
                [
                    0.0,
                    -0.4444444444444439,
                    2.6666666666666656,
                    0.0,
                    -2.2222222222222214,
                ],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        # We test only the mass-Hdiv part
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    def test_dual_hybrid_vem_3d_iso_cart(self):
        g = pp.CartGrid([2, 2, 2], [1, 1, 1])
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["neu"])

        solver = hybrid.HybridDualVEM(keyword="flow")

        data = make_dictionary(g, perm, bc)
        M = solver.matrix_rhs(g, data)[0].todense()

        # np.savetxt('matrix.txt', M, delimiter=',', newline='],\n[')
        M_known = matrix_for_test_dual_hybrid_vem_3d_iso_cart()

        rtol = 1e-14
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    def test_dual_hybrid_vem_3d_ani_cart(self):
        g = pp.CartGrid([2, 2, 2], [1, 1, 1])
        g.compute_geometry()

        kxx = np.square(g.cell_centers[1, :]) + 1
        kyy = np.square(g.cell_centers[0, :]) + 1
        kzz = g.cell_centers[2, :] + 1
        kxy = -np.multiply(g.cell_centers[0, :], g.cell_centers[1, :])
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kyy, kxy=kxy, kzz=kzz)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["neu"])

        solver = hybrid.HybridDualVEM(keyword="flow")

        data = make_dictionary(g, perm, bc)
        M = solver.matrix_rhs(g, data)[0].todense()
        # np.savetxt('matrix.txt', M, delimiter=',', newline='],\n[')
        M_known = matrix_for_test_dual_hybrid_vem_3d_ani_cart()

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    def test_dual_vem_2d_iso_cart_surf(self):
        g = pp.CartGrid([2, 1], [1, 1])
        R = pp.map_geometry.rotation_matrix(np.pi / 4.0, [0, 1, 0])
        g.nodes = np.dot(R, g.nodes)
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx)
        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["neu"])

        solver = hybrid.HybridDualVEM(keyword="flow")

        data = make_dictionary(g, perm, bc)
        M = solver.matrix_rhs(g, data)[0].todense()
        M_known = np.array(
            [
                [-2.25, 1.75, 0.0, 0.25, 0.0, 0.25, 0.0],
                [1.75, -4.5, 1.75, 0.25, 0.25, 0.25, 0.25],
                [0.0, 1.75, -2.25, 0.0, 0.25, 0.0, 0.25],
                [0.25, 0.25, 0.0, -0.75, 0.0, 0.25, 0.0],
                [0.0, 0.25, 0.25, 0.0, -0.75, 0.0, 0.25],
                [0.25, 0.25, 0.0, 0.25, 0.0, -0.75, 0.0],
                [0.0, 0.25, 0.25, 0.0, 0.25, 0.0, -0.75],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    def test_dual_vem_1d_iso_line(self):
        g = pp.CartGrid(3, 1)
        R = pp.map_geometry.rotation_matrix(np.pi / 6.0, [0, 0, 1])
        g.nodes = np.dot(R, g.nodes)
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["neu"])

        solver = hybrid.HybridDualVEM(keyword="flow")

        data = make_dictionary(g, perm, bc)
        M = solver.matrix_rhs(g, data)[0].todense()
        M_known = np.array(
            [
                [-3.0, 3.0, 0.0, 0.0],
                [3.0, -6.0, 3.0, 0.0],
                [0.0, 3.0, -6.0, 3.0],
                [0.0, 0.0, 3.0, -3.0],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))


def matrix_for_test_dual_hybrid_vem_3d_iso_cart():
    return np.array(
        [
            [
                -7.886751345948128655e-01,
                2.113248654051872177e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                2.113248654051871900e-01,
                -1.577350269189625731e00,
                2.113248654051872177e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                2.113248654051871900e-01,
                -7.886751345948129766e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948124214e-01,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                -1.577350269189624843e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                -7.886751345948125325e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948128655e-01,
                2.113248654051872177e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051871900e-01,
                -1.577350269189625731e00,
                2.113248654051872177e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051871900e-01,
                -7.886751345948129766e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948124214e-01,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                -1.577350269189624843e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064328e-01,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                -7.886751345948125325e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
            ],
            [
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948128655e-01,
                0.000000000000000000e00,
                2.113248654051872177e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948128655e-01,
                0.000000000000000000e00,
                2.113248654051872177e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051871900e-01,
                0.000000000000000000e00,
                -1.577350269189625287e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051871900e-01,
                0.000000000000000000e00,
                -1.577350269189625287e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                -7.886751345948125325e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                -7.886751345948125325e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948128655e-01,
                0.000000000000000000e00,
                2.113248654051872177e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948128655e-01,
                0.000000000000000000e00,
                2.113248654051872177e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051871900e-01,
                0.000000000000000000e00,
                -1.577350269189625287e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051871900e-01,
                0.000000000000000000e00,
                -1.577350269189625287e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                -7.886751345948125325e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                -7.886751345948125325e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
            ],
            [
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948128655e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051872177e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948128655e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051872177e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948124214e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948124214e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051871900e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.577350269189625731e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051872177e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051871900e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.577350269189625731e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051872177e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.577350269189624843e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.577350269189624843e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051871900e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948129766e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051871900e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948129766e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948125325e-01,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.113248654051867459e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.886751345948125325e-01,
            ],
        ]
    )


def matrix_for_test_dual_hybrid_vem_3d_ani_cart():
    return np.array(
        [
            [
                -8.199251345948128655e-01,
                2.425748654051871900e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.755875672974064328e-01,
                0.000000000000000000e00,
                1.130875672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064883e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                2.425748654051871900e-01,
                -1.619230616718567273e00,
                2.631945178762451487e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.130875672974064189e-01,
                2.277777410618773146e-01,
                1.755875672974064328e-01,
                4.027774106187737013e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                1.340277410618773424e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                2.631945178762451487e-01,
                -7.993054821237547403e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                4.027774106187746728e-02,
                0.000000000000000000e00,
                2.277777410618773146e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.049305482123754185e00,
                5.131945178762447046e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.277777410618773146e-01,
                0.000000000000000000e00,
                4.027774106187742564e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.131945178762448156e-01,
                -2.119230616718566829e00,
                4.925748654051868014e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                4.027774106187743952e-02,
                4.255875672974062662e-01,
                2.277777410618773146e-01,
                -1.369124327025934840e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                4.925748654051868014e-01,
                -1.069925134594812643e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.369124327025934285e-01,
                0.000000000000000000e00,
                4.255875672974062662e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -8.199251345948128655e-01,
                2.425748654051871900e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.755875672974064328e-01,
                0.000000000000000000e00,
                1.130875672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.425748654051871900e-01,
                -1.619230616718567273e00,
                2.631945178762451487e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.130875672974064189e-01,
                2.277777410618773146e-01,
                1.755875672974064328e-01,
                4.027774106187737013e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.340277410618773146e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.631945178762451487e-01,
                -7.993054821237547403e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                4.027774106187746728e-02,
                0.000000000000000000e00,
                2.277777410618773146e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.049305482123754185e00,
                5.131945178762447046e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.277777410618773146e-01,
                0.000000000000000000e00,
                4.027774106187742564e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.131945178762448156e-01,
                -2.119230616718566829e00,
                4.925748654051868014e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                4.027774106187743952e-02,
                4.255875672974062662e-01,
                2.277777410618773146e-01,
                -1.369124327025934840e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                1.443375672974064050e-01,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                4.925748654051868014e-01,
                -1.069925134594812643e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.369124327025934285e-01,
                0.000000000000000000e00,
                4.255875672974062662e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
            ],
            [
                1.755875672974064328e-01,
                1.130875672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -8.199251345948127545e-01,
                0.000000000000000000e00,
                2.425748654051871900e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                2.277777410618773146e-01,
                4.027774106187749503e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.049305482123754629e00,
                0.000000000000000000e00,
                5.131945178762450377e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                1.130875672974064744e-01,
                1.755875672974064605e-01,
                0.000000000000000000e00,
                2.277777410618772869e-01,
                4.027774106187745340e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.425748654051871900e-01,
                0.000000000000000000e00,
                -1.619230616718567273e00,
                0.000000000000000000e00,
                2.631945178762449267e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064883e-01,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                4.027774106187741177e-02,
                2.277777410618772869e-01,
                0.000000000000000000e00,
                4.255875672974062107e-01,
                -1.369124327025934007e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.131945178762450377e-01,
                0.000000000000000000e00,
                -2.119230616718566829e00,
                0.000000000000000000e00,
                4.925748654051864683e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                4.027774106187745340e-02,
                2.277777410618772869e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.631945178762449267e-01,
                0.000000000000000000e00,
                -7.993054821237544072e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.369124327025934562e-01,
                4.255875672974062662e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                4.925748654051864683e-01,
                0.000000000000000000e00,
                -1.069925134594812199e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.755875672974064328e-01,
                1.130875672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -8.199251345948127545e-01,
                0.000000000000000000e00,
                2.425748654051871900e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.277777410618773146e-01,
                4.027774106187749503e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.049305482123754629e00,
                0.000000000000000000e00,
                5.131945178762450377e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.130875672974064744e-01,
                1.755875672974064605e-01,
                0.000000000000000000e00,
                2.277777410618772869e-01,
                4.027774106187745340e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.425748654051871900e-01,
                0.000000000000000000e00,
                -1.619230616718567273e00,
                0.000000000000000000e00,
                2.631945178762449267e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                4.027774106187741177e-02,
                2.277777410618772869e-01,
                0.000000000000000000e00,
                4.255875672974062107e-01,
                -1.369124327025934007e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.131945178762450377e-01,
                0.000000000000000000e00,
                -2.119230616718566829e00,
                0.000000000000000000e00,
                4.925748654051864683e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                1.340277410618773424e-01,
                0.000000000000000000e00,
                1.443375672974064050e-01,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                4.027774106187745340e-02,
                2.277777410618772869e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.631945178762449267e-01,
                0.000000000000000000e00,
                -7.993054821237544072e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.369124327025934562e-01,
                4.255875672974062662e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                4.925748654051864683e-01,
                0.000000000000000000e00,
                -1.069925134594812199e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
            ],
            [
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -9.136751345948124214e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                3.363248654051868014e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                1.340277410618773701e-01,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -8.930554821237545182e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                3.569445178762449267e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -8.930554821237545182e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                3.569445178762449822e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -9.136751345948125325e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                3.363248654051868569e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                3.363248654051868014e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.077350269189624843e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.863248654051869124e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                1.340277410618773701e-01,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                3.569445178762449267e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.036110964247509258e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                6.069445178762449267e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                1.340277410618774256e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                3.569445178762449822e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.036110964247508814e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                6.069445178762449267e-01,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974063773e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974063773e-01,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                3.363248654051868569e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.077350269189625287e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.863248654051868014e-01,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                1.443375672974064605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.863248654051869124e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.163675134594812421e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773424e-01,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                1.340277410618773424e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                6.069445178762449267e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.143055482123754407e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773701e-01,
                1.340277410618773701e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
                1.340277410618773979e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                6.069445178762449267e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.143055482123754407e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.443375672974064050e-01,
                0.000000000000000000e00,
                1.443375672974064328e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.863248654051869124e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.163675134594812643e00,
            ],
        ]
    )


if __name__ == "__main__":
    unittest.main()
