import numpy as np
import scipy.sparse as sps
import unittest

import porepy as pp


class BasicsTest(unittest.TestCase):
    def _matrix(self, g, perm, bc):
        solver = pp.MVEM(keyword="flow")

        data = pp.initialize_default_data(
            g, {}, "flow", {"second_order_tensor": perm, "bc": bc}
        )
        solver.discretize(g, data)

        return solver.assemble_matrix(g, data).todense()

    # ------------------------------------------------------------------------------#

    def test_dual_vem_1d_iso(self):
        g = pp.CartGrid(3, 1)
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx)
        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)

        M_known = 1e-2 * np.array(
            [
                [25, -(8 + 1 / 3.0), 0, 0, 1e2, 0, 0],
                [-(8 + 1 / 3.0), 50, -(8 + 1 / 3.0), 0, -1e2, 1e2, 0],
                [0, -(8 + 1 / 3.0), 50, -(8 + 1 / 3.0), 0, -1e2, 1e2],
                [0, 0, -(8 + 1 / 3.0), 25, 0, 0, -1e2],
                [1e2, -1e2, 0, 0, 0, 0, 0],
                [0, 1e2, -1e2, 0, 0, 0, 0],
                [0, 0, 1e2, -1e2, 0, 0, 0],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    # ------------------------------------------------------------------------------#

    def test_dual_vem_1d_ani(self):
        g = pp.CartGrid(3, 1)
        g.compute_geometry()

        kxx = np.sin(g.cell_centers[0, :]) + 1
        perm = pp.SecondOrderTensor(
            kxx, kyy=np.ones(g.num_cells), kzz=np.ones(g.num_cells)
        )
        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)
        M_known = 1e-2 * np.array(
            [
                [21.4427334468001192, -7.14757781560004, 0, 0, 1e2, 0, 0],
                [
                    -7.14757781560004,
                    38.3411844655562319,
                    -5.6328170062520355,
                    0,
                    -1e2,
                    1e2,
                    0,
                ],
                [
                    0,
                    -5.6328170062520355,
                    31.2648069176318977,
                    -4.788785299625264,
                    0,
                    -1e2,
                    1e2,
                ],
                [0, 0, -4.7887852996252649, 14.3663558988757991, 0, 0, -1e2],
                [1e2, -1e2, 0, 0, 0, 0, 0],
                [0, 1e2, -1e2, 0, 0, 0, 0],
                [0, 0, 1e2, -1e2, 0, 0, 0],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    # ------------------------------------------------------------------------------#

    def test_dual_vem_2d_iso_cart(self):
        g = pp.CartGrid([2, 1], [1, 1])
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kxx, kzz=kxx)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)
        # Matrix computed with an already validated code (MRST)
        M_known = np.array(
            [
                [0.625, -0.375, 0, 0, 0, 0, 0, 1, 0],
                [-0.375, 1.25, -0.375, 0, 0, 0, 0, -1, 1],
                [0, -0.375, 0.625, 0, 0, 0, 0, 0, -1],
                [0, 0, 0, 1, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 1, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 1, 0, -1, 0],
                [0, 0, 0, 0, 0, 0, 1, 0, -1],
                [1, -1, 0, 1, 0, -1, 0, 0, 0],
                [0, 1, -1, 0, 1, 0, -1, 0, 0],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    # ------------------------------------------------------------------------------#

    def test_dual_vem_2d_ani_cart(self):
        g = pp.CartGrid([2, 1], [1, 1])
        g.compute_geometry()

        kxx = np.square(g.cell_centers[1, :]) + 1
        kyy = np.square(g.cell_centers[0, :]) + 1
        kxy = -np.multiply(g.cell_centers[0, :], g.cell_centers[1, :])
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kyy, kxy=kxy)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)
        # Matrix computed with an already validated code (MRST)
        M_known = np.array(
            [
                [
                    0.625000000000000,
                    -0.422619047619048,
                    0,
                    0.023809523809524,
                    0,
                    0.023809523809524,
                    0,
                    1,
                    0,
                ],
                [
                    -0.422619047619048,
                    1.267241379310345,
                    -0.426724137931035,
                    0.023809523809524,
                    0.051724137931034,
                    0.023809523809524,
                    0.051724137931034,
                    -1,
                    1,
                ],
                [
                    0,
                    -0.426724137931035,
                    0.642241379310345,
                    0,
                    0.051724137931034,
                    0,
                    0.051724137931034,
                    0,
                    -1,
                ],
                [
                    0.023809523809524,
                    0.023809523809524,
                    0,
                    1,
                    0,
                    -0.047619047619048,
                    0,
                    1,
                    0,
                ],
                [
                    0,
                    0.051724137931034,
                    0.051724137931034,
                    0,
                    0.879310344827586,
                    0,
                    -0.189655172413793,
                    0,
                    1,
                ],
                [
                    0.023809523809524,
                    0.023809523809524,
                    0,
                    -0.047619047619048,
                    0,
                    1,
                    0,
                    -1,
                    0,
                ],
                [
                    0,
                    0.051724137931034,
                    0.051724137931034,
                    0,
                    -0.189655172413793,
                    0,
                    0.879310344827586,
                    0,
                    -1,
                ],
                [1, -1, 0, 1, 0, -1, 0, 0, 0],
                [0, 1, -1, 0, 1, 0, -1, 0, 0],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    # ------------------------------------------------------------------------------#

    def test_dual_vem_2d_iso_simplex(self):
        g = pp.StructuredTriangleGrid([1, 1], [1, 1])
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kxx)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)
        # Matrix computed with an already validated code (MRST)
        faces = np.arange(5)
        map_faces = np.array([1, 4, 0, 2, 3])
        M_known = np.array(
            [
                [
                    0.888888888888889,
                    0.277777777777778,
                    0.277777777777778,
                    -0.277777777777778,
                    -0.277777777777778,
                    -1.000000000000000,
                    1,
                ],
                [0.277777777777778, 0.611111111111111, 0.111111111111111, 0, 0, -1, 0],
                [0.277777777777778, 0.111111111111111, 0.611111111111111, 0, 0, -1, 0],
                [-0.277777777777778, 0, 0, 0.611111111111111, 0.111111111111111, 0, -1],
                [-0.277777777777778, 0, 0, 0.111111111111111, 0.611111111111111, 0, 1],
                [-1, -1, -1, 0, 0, 0, 0],
                [1, 0, 0, -1, -1, 0, 0],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        # We test only the mass-Hdiv part
        self.assertTrue(
            np.allclose(
                M[np.ix_(faces, faces)],
                M_known[np.ix_(map_faces, map_faces)],
                rtol,
                atol,
            )
        )

    # ------------------------------------------------------------------------------#

    def test_dual_vem_2d_iso_simplex_mixed_bc(self):
        g = pp.StructuredTriangleGrid([2, 2], [1, 1])
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kxx)

        tol = 1e-6
        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bf_centers = g.face_centers[:, bf]
        left = bf_centers[0, :] < 0 + tol
        right = bf_centers[0, :] > 1 - tol

        labels = np.array(["neu"] * bf.size)
        labels[left] = "rob"
        labels[right] = "dir"

        bc = pp.BoundaryCondition(g, bf, labels)
        bc.robin_weight[bf[left]] = 2

        bc_val = np.zeros(g.num_faces)
        bc_val[bf[left]] = 3

        solver = pp.MVEM(keyword="flow")

        data = pp.initialize_default_data(
            g, {}, "flow", {"second_order_tensor": perm, "bc": bc, "bc_values": bc_val}
        )
        solver.discretize(g, data)
        M, rhs = solver.assemble_matrix_rhs(g, data)
        up = sps.linalg.spsolve(M, rhs)

        p = solver.extract_pressure(g, up, data)
        u = solver.extract_flux(g, up, data)
        P0u = solver.project_flux(g, u, data)

        p_ex = 1 - g.cell_centers[0, :]
        P0u_ex = np.vstack(
            (np.ones(g.num_cells), np.zeros(g.num_cells), np.zeros(g.num_cells))
        )

        self.assertTrue(np.allclose(p, p_ex))
        self.assertTrue(np.allclose(P0u, P0u_ex))

        # np.savetxt('matrix.txt', M.todense(), delimiter=',', newline='],\n[')
        M_known = matrix_for_test_dual_vem_2d_iso_simplex_mixed_bc()

        self.assertTrue(np.allclose(M.todense(), M_known))

    # ------------------------------------------------------------------------------#

    def test_dual_vem_2d_ani_simplex(self):
        g = pp.StructuredTriangleGrid([1, 1], [1, 1])
        g.compute_geometry()

        kxx = np.square(g.cell_centers[1, :]) + 1
        kyy = np.square(g.cell_centers[0, :]) + 1
        kxy = -np.multiply(g.cell_centers[0, :], g.cell_centers[1, :])
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kyy, kxy=kxy, kzz=1)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)
        # Matrix computed with an already validated code (MRST)
        faces = np.arange(5)
        map_faces = np.array([1, 4, 0, 2, 3])
        M_known = np.array(
            [
                [
                    0.865079365079365,
                    0.337301587301587,
                    0.301587301587302,
                    -0.301587301587302,
                    -0.337301587301587,
                    -1,
                    1,
                ],
                [0.337301587301587, 0.599206349206349, 0.134920634920635, 0, 0, -1, 0],
                [0.301587301587302, 0.134920634920635, 0.634920634920635, 0, 0, -1, 0],
                [-0.301587301587302, 0, 0, 0.634920634920635, 0.134920634920635, 0, -1],
                [-0.337301587301587, 0, 0, 0.134920634920635, 0.599206349206349, 0, -1],
                [-1, -1, -1, 0, 0, 0, 0],
                [1, 0, 0, -1, -1, 0, 0],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        # We test only the mass-Hdiv part
        self.assertTrue(
            np.allclose(
                M[np.ix_(faces, faces)],
                M_known[np.ix_(map_faces, map_faces)],
                rtol,
                atol,
            )
        )

    # ------------------------------------------------------------------------------#

    def test_dual_vem_3d_iso_cart(self):
        g = pp.CartGrid([2, 2, 2], [1, 1, 1])
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kxx, kzz=kxx)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)
        # np.savetxt('matrix.txt', M, delimiter=',', newline='],\n[')
        M_known = matrix_for_test_dual_vem_3d_iso_cart()

        rtol = 1e-14
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    # ------------------------------------------------------------------------------#

    def test_dual_vem_3d_ani_cart(self):
        g = pp.CartGrid([2, 2, 2], [1, 1, 1])
        g.compute_geometry()

        kxx = np.square(g.cell_centers[1, :]) + 1
        kyy = np.square(g.cell_centers[0, :]) + 1
        kzz = g.cell_centers[2, :] + 1
        kxy = -np.multiply(g.cell_centers[0, :], g.cell_centers[1, :])
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kyy, kxy=kxy, kzz=kzz)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)
        # np.savetxt('matrix.txt', M, delimiter=',', newline='],\n[')
        M_known = matrix_for_test_dual_vem_3d_ani_cart()

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    # ------------------------------------------------------------------------------#

    def test_dual_vem_1d_iso_line(self):
        g = pp.CartGrid(3, 1)
        R = pp.map_geometry.rotation_matrix(np.pi / 6.0, [0, 0, 1])
        g.nodes = np.dot(R, g.nodes)
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx, kyy=1, kzz=1)
        perm.rotate(R)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)
        # Matrix computed with an already validated code (MRST)
        M_known = 1e-2 * np.array(
            [
                [25, -(8 + 1 / 3.0), 0, 0, 1e2, 0, 0],
                [-(8 + 1 / 3.0), 50, -(8 + 1 / 3.0), 0, -1e2, 1e2, 0],
                [0, -(8 + 1 / 3.0), 50, -(8 + 1 / 3.0), 0, -1e2, 1e2],
                [0, 0, -(8 + 1 / 3.0), 25, 0, 0, -1e2],
                [1e2, -1e2, 0, 0, 0, 0, 0],
                [0, 1e2, -1e2, 0, 0, 0, 0],
                [0, 0, 1e2, -1e2, 0, 0, 0],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    # ------------------------------------------------------------------------------#

    def test_dual_vem_2d_iso_cart_surf(self):
        g = pp.CartGrid([2, 1], [1, 1])
        R = pp.map_geometry.rotation_matrix(np.pi / 4.0, [0, 1, 0])
        g.nodes = np.dot(R, g.nodes)
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kxx, kzz=1)
        perm.rotate(R)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)
        # Matrix computed with an already validated code (MRST)
        M_known = np.array(
            [
                [0.625, -0.375, 0, 0, 0, 0, 0, 1, 0],
                [-0.375, 1.25, -0.375, 0, 0, 0, 0, -1, 1],
                [0, -0.375, 0.625, 0, 0, 0, 0, 0, -1],
                [0, 0, 0, 1, 0, 0, 0, 1, 0],
                [0, 0, 0, 0, 1, 0, 0, 0, 1],
                [0, 0, 0, 0, 0, 1, 0, -1, 0],
                [0, 0, 0, 0, 0, 0, 1, 0, -1],
                [1, -1, 0, 1, 0, -1, 0, 0, 0],
                [0, 1, -1, 0, 1, 0, -1, 0, 0],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    # ------------------------------------------------------------------------------#

    def test_dual_vem_2d_ani_cart_surf(self):
        g = pp.CartGrid([2, 1], [1, 1])
        g.compute_geometry()

        kxx = np.square(g.cell_centers[1, :]) + 1
        kyy = np.square(g.cell_centers[0, :]) + 1
        kxy = -np.multiply(g.cell_centers[0, :], g.cell_centers[1, :])
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kyy, kxy=kxy, kzz=1)

        R = pp.map_geometry.rotation_matrix(np.pi / 3.0, [1, 1, 0])
        perm.rotate(R)
        g.nodes = np.dot(R, g.nodes)
        g.compute_geometry()

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)
        # Matrix computed with an already validated code (MRST)
        M_known = np.array(
            [
                [
                    0.625000000000000,
                    -0.422619047619048,
                    0,
                    0.023809523809524,
                    0,
                    0.023809523809524,
                    0,
                    1,
                    0,
                ],
                [
                    -0.422619047619048,
                    1.267241379310345,
                    -0.426724137931035,
                    0.023809523809524,
                    0.051724137931034,
                    0.023809523809524,
                    0.051724137931034,
                    -1,
                    1,
                ],
                [
                    0,
                    -0.426724137931035,
                    0.642241379310345,
                    0,
                    0.051724137931034,
                    0,
                    0.051724137931034,
                    0,
                    -1,
                ],
                [
                    0.023809523809524,
                    0.023809523809524,
                    0,
                    1,
                    0,
                    -0.047619047619048,
                    0,
                    1,
                    0,
                ],
                [
                    0,
                    0.051724137931034,
                    0.051724137931034,
                    0,
                    0.879310344827586,
                    0,
                    -0.189655172413793,
                    0,
                    1,
                ],
                [
                    0.023809523809524,
                    0.023809523809524,
                    0,
                    -0.047619047619048,
                    0,
                    1,
                    0,
                    -1,
                    0,
                ],
                [
                    0,
                    0.051724137931034,
                    0.051724137931034,
                    0,
                    -0.189655172413793,
                    0,
                    0.879310344827586,
                    0,
                    -1,
                ],
                [1, -1, 0, 1, 0, -1, 0, 0, 0],
                [0, 1, -1, 0, 1, 0, -1, 0, 0],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        self.assertTrue(np.allclose(M, M_known, rtol, atol))

    # ------------------------------------------------------------------------------#

    def test_dual_vem_2d_iso_simplex_surf(self):
        g = pp.StructuredTriangleGrid([1, 1], [1, 1])
        R = pp.map_geometry.rotation_matrix(-np.pi / 4.0, [1, 1, -1])
        g.nodes = np.dot(R, g.nodes)
        g.compute_geometry()

        kxx = np.ones(g.num_cells)
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kxx, kzz=1)
        perm.rotate(R)

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)
        # Matrix computed with an already validated code (MRST)
        faces = np.arange(5)
        map_faces = np.array([1, 4, 0, 2, 3])
        M_known = np.array(
            [
                [
                    0.888888888888889,
                    0.277777777777778,
                    0.277777777777778,
                    -0.277777777777778,
                    -0.277777777777778,
                    -1.000000000000000,
                    1,
                ],
                [0.277777777777778, 0.611111111111111, 0.111111111111111, 0, 0, -1, 0],
                [0.277777777777778, 0.111111111111111, 0.611111111111111, 0, 0, -1, 0],
                [-0.277777777777778, 0, 0, 0.611111111111111, 0.111111111111111, 0, -1],
                [-0.277777777777778, 0, 0, 0.111111111111111, 0.611111111111111, 0, 1],
                [-1, -1, -1, 0, 0, 0, 0],
                [1, 0, 0, -1, -1, 0, 0],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        # We test only the mass-Hdiv part
        self.assertTrue(
            np.allclose(
                M[np.ix_(faces, faces)],
                M_known[np.ix_(map_faces, map_faces)],
                rtol,
                atol,
            )
        )

    # ------------------------------------------------------------------------------#

    def test_dual_vem_2d_ani_simplex_surf(self):
        g = pp.StructuredTriangleGrid([1, 1], [1, 1])
        g.compute_geometry()

        kxx = np.square(g.cell_centers[1, :]) + 1
        kyy = np.square(g.cell_centers[0, :]) + 1
        kxy = -np.multiply(g.cell_centers[0, :], g.cell_centers[1, :])
        perm = pp.SecondOrderTensor(kxx=kxx, kyy=kyy, kxy=kxy, kzz=1)

        R = pp.map_geometry.rotation_matrix(np.pi / 3.0, [1, 1, 0])
        perm.rotate(R)
        g.nodes = np.dot(R, g.nodes)
        g.compute_geometry()

        bf = g.tags["domain_boundary_faces"].nonzero()[0]
        bc = pp.BoundaryCondition(g, bf, bf.size * ["dir"])

        M = self._matrix(g, perm, bc)
        # assemble_matrix_rhs computed with an already validated code (MRST)
        faces = np.arange(5)
        map_faces = np.array([1, 4, 0, 2, 3])
        M_known = np.array(
            [
                [
                    0.865079365079365,
                    0.337301587301587,
                    0.301587301587302,
                    -0.301587301587302,
                    -0.337301587301587,
                    -1,
                    1,
                ],
                [0.337301587301587, 0.599206349206349, 0.134920634920635, 0, 0, -1, 0],
                [0.301587301587302, 0.134920634920635, 0.634920634920635, 0, 0, -1, 0],
                [-0.301587301587302, 0, 0, 0.634920634920635, 0.134920634920635, 0, -1],
                [-0.337301587301587, 0, 0, 0.134920634920635, 0.599206349206349, 0, -1],
                [-1, -1, -1, 0, 0, 0, 0],
                [1, 0, 0, -1, -1, 0, 0],
            ]
        )

        rtol = 1e-15
        atol = rtol
        self.assertTrue(np.allclose(M, M.T, rtol, atol))
        # We test only the mass-Hdiv part
        self.assertTrue(
            np.allclose(
                M[np.ix_(faces, faces)],
                M_known[np.ix_(map_faces, map_faces)],
                rtol,
                atol,
            )
        )


# ------------------------------------------------------------------------------#


def matrix_for_test_dual_vem_2d_iso_simplex_mixed_bc():
    return np.array(
        [
            [
                2.000000000000000444e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                1.611111111111111160e00,
                -2.777777777777777901e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.111111111111110494e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                2.777777777777777901e-01,
                -2.777777777777777901e-01,
                8.888888888888890616e-01,
                0.000000000000000000e00,
                2.777777777777777901e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.777777777777776791e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.000000000000000444e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                1.111111111111110494e-01,
                0.000000000000000000e00,
                2.777777777777777901e-01,
                0.000000000000000000e00,
                1.222222222222222099e00,
                2.777777777777777901e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.111111111111110494e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.777777777777777901e-01,
                2.777777777777777901e-01,
                8.888888888888890616e-01,
                2.777777777777777901e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.777777777777778456e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.111111111111110494e-01,
                0.000000000000000000e00,
                2.777777777777777901e-01,
                6.111111111111110494e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                1.111111111111110494e-01,
                -2.777777777777776791e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.222222222222221877e00,
                0.000000000000000000e00,
                -2.777777777777776791e-01,
                0.000000000000000000e00,
                -1.111111111111110494e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.611111111111111160e00,
                -2.777777777777777901e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.111111111111110494e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.777777777777776791e-01,
                -2.777777777777777901e-01,
                8.888888888888889506e-01,
                0.000000000000000000e00,
                2.777777777777777901e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.777777777777776791e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.111111111111110494e-01,
                -2.777777777777778456e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.222222222222222321e00,
                0.000000000000000000e00,
                -2.777777777777777901e-01,
                -1.111111111111110217e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.111111111111110494e-01,
                0.000000000000000000e00,
                2.777777777777777901e-01,
                0.000000000000000000e00,
                1.222222222222222321e00,
                2.777777777777777901e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.111111111111110217e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.777777777777777901e-01,
                2.777777777777777901e-01,
                8.888888888888890616e-01,
                2.777777777777777901e-01,
                0.000000000000000000e00,
                -2.777777777777777901e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.111111111111110217e-01,
                0.000000000000000000e00,
                2.777777777777777901e-01,
                6.111111111111111605e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.000000000000000444e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.000000000000000444e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                -1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
        ]
    )


# ------------------------------------------------------------------------------#


def matrix_for_test_dual_vem_3d_iso_cart():
    return np.array(
        [
            [
                1.077350269189625731e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                -7.735026918962584208e-02,
                2.154700538379251462e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                1.077350269189625731e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189626175e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                2.154700538379252350e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                1.077350269189626175e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189625731e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                2.154700538379251462e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                1.077350269189625731e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189626175e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                2.154700538379252350e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                1.077350269189626175e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189625731e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189625731e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                2.154700538379251906e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                2.154700538379251906e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                1.077350269189626175e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                1.077350269189626175e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189625731e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189625731e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                2.154700538379251906e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                2.154700538379251906e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                1.077350269189626175e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                1.077350269189626175e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189625731e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189625731e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189626175e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189626175e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.154700538379251462e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.154700538379251462e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.154700538379252350e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.154700538379252350e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189625731e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962584208e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189625731e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189626175e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -7.735026918962539799e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.077350269189626175e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
            ],
            [
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
        ]
    )


# ------------------------------------------------------------------------------#


def matrix_for_test_dual_vem_3d_ani_cart():
    return np.array(
        [
            [
                1.049572491411848052e00,
                -1.051280469674036322e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.777777777777776930e-02,
                0.000000000000000000e00,
                2.777777777777776930e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                -1.051280469674036322e-01,
                2.152103550539137267e00,
                -1.409925975888278171e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.777777777777776930e-02,
                5.769230769230768163e-02,
                2.777777777777776930e-02,
                5.769230769230768163e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                -1.409925975888278171e-01,
                1.102531059127289437e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.769230769230768163e-02,
                0.000000000000000000e00,
                5.769230769230768163e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                9.486849052811356930e-01,
                -2.948387514349815053e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.769230769230771633e-02,
                0.000000000000000000e00,
                5.769230769230771633e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.948387514349815053e-01,
                1.893682233294291084e00,
                -2.097032103660962377e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.769230769230771633e-02,
                1.323529411764707009e-01,
                5.769230769230771633e-02,
                1.323529411764707009e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.097032103660962377e-01,
                9.449973280131553910e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.323529411764707009e-01,
                0.000000000000000000e00,
                1.323529411764707009e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.049572491411848052e00,
                -1.051280469674036322e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.777777777777776930e-02,
                0.000000000000000000e00,
                2.777777777777776930e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.051280469674036322e-01,
                2.152103550539137267e00,
                -1.409925975888278171e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.777777777777776930e-02,
                5.769230769230768163e-02,
                2.777777777777776930e-02,
                5.769230769230768163e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.409925975888278171e-01,
                1.102531059127289437e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.769230769230768163e-02,
                0.000000000000000000e00,
                5.769230769230768163e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                9.486849052811356930e-01,
                -2.948387514349815053e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.769230769230771633e-02,
                0.000000000000000000e00,
                5.769230769230771633e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.948387514349815053e-01,
                1.893682233294291084e00,
                -2.097032103660962377e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.769230769230771633e-02,
                1.323529411764707009e-01,
                5.769230769230771633e-02,
                1.323529411764707009e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                1.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.097032103660962377e-01,
                9.449973280131553910e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.323529411764707009e-01,
                0.000000000000000000e00,
                1.323529411764707009e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
            ],
            [
                2.777777777777776930e-02,
                2.777777777777776930e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.049572491411848052e00,
                0.000000000000000000e00,
                -1.051280469674036322e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                5.769230769230767469e-02,
                5.769230769230767469e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                9.486849052811355820e-01,
                0.000000000000000000e00,
                -2.948387514349816163e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                2.777777777777776930e-02,
                2.777777777777776930e-02,
                0.000000000000000000e00,
                5.769230769230770939e-02,
                5.769230769230770939e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.051280469674036322e-01,
                0.000000000000000000e00,
                2.152103550539137711e00,
                0.000000000000000000e00,
                -1.409925975888275951e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                5.769230769230767469e-02,
                5.769230769230767469e-02,
                0.000000000000000000e00,
                1.323529411764707009e-01,
                1.323529411764707009e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.948387514349816163e-01,
                0.000000000000000000e00,
                1.893682233294291084e00,
                0.000000000000000000e00,
                -2.097032103660961266e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.769230769230770939e-02,
                5.769230769230770939e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.409925975888275951e-01,
                0.000000000000000000e00,
                1.102531059127289659e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.323529411764707009e-01,
                1.323529411764707009e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.097032103660961266e-01,
                0.000000000000000000e00,
                9.449973280131556130e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.777777777777776930e-02,
                2.777777777777776930e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.049572491411848052e00,
                0.000000000000000000e00,
                -1.051280469674036322e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.769230769230767469e-02,
                5.769230769230767469e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                9.486849052811355820e-01,
                0.000000000000000000e00,
                -2.948387514349816163e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                2.777777777777776930e-02,
                2.777777777777776930e-02,
                0.000000000000000000e00,
                5.769230769230770939e-02,
                5.769230769230770939e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.051280469674036322e-01,
                0.000000000000000000e00,
                2.152103550539137711e00,
                0.000000000000000000e00,
                -1.409925975888275951e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.769230769230767469e-02,
                5.769230769230767469e-02,
                0.000000000000000000e00,
                1.323529411764707009e-01,
                1.323529411764707009e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.948387514349816163e-01,
                0.000000000000000000e00,
                1.893682233294291084e00,
                0.000000000000000000e00,
                -2.097032103660961266e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                5.769230769230770939e-02,
                5.769230769230770939e-02,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.409925975888275951e-01,
                0.000000000000000000e00,
                1.102531059127289659e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.323529411764707009e-01,
                1.323529411764707009e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.097032103660961266e-01,
                0.000000000000000000e00,
                9.449973280131556130e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                9.773502691896260863e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.773502691896255978e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.021761828358058732e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.217618283580583549e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.021761828358058732e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.217618283580584104e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                9.773502691896260863e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.773502691896256533e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.773502691896255978e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.840414824093537849e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.916359834753400326e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.217618283580583549e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.929237942430403141e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -3.360475426437727897e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.217618283580584104e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.929237942430403141e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -3.360475426437727897e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.773502691896256533e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.840414824093537849e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.916359834753400326e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.916359834753400326e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                8.630645549039116515e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -3.360475426437727897e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                9.074761140723444086e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -3.360475426437727897e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                9.074761140723444086e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -2.916359834753400326e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                8.630645549039116515e-01,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
            ],
            [
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
            [
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                -1.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
                0.000000000000000000e00,
            ],
        ]
    )


# ------------------------------------------------------------------------------#
if __name__ == "__main__":
    unittest.main()
